// online-mfcnet.cc
// If you want to put this file under the directory, it file is placed in under the "feat" directory.

#ifdef _MSC_VER
#define NOMINMAX
#endif
#include "online-mfcnet.h"

namespace kaldi {

OnlineFeatureMfcnet::OnlineFeatureMfcnet(const Poco::Net::StreamSocket& socket)
  : socket_(socket), feat_dim_(0), begin_frame_(0), num_frames_(0), frame_size_(0), input_finished_(false)
{
  {
    kaldi::int32 srcinfo_size;
    while(!(socket_.available() >= sizeof(srcinfo_size))){
#ifdef _MSC_VER
      ::Sleep((DWORD)1);
#else
      usleep(1000);
#endif
    }
    socket_.receiveBytes(&srcinfo_size, sizeof(srcinfo_size));
    KALDI_ASSERT(srcinfo_size == sizeof(source_info_)); // SourceInfo header must be 28 Bytes
  }

  {
    while(!(socket_.available() >= sizeof(source_info_))){
#ifdef _MSC_VER
      ::Sleep((DWORD)1);
#else
      usleep(1000);
#endif
    }
    socket_.receiveBytes(&source_info_, sizeof(source_info_));
    KALDI_LOG << "source_id = " << source_info_.source_id
              << ", azimuth = " << source_info_.azimuth
              << ", elevation = " << source_info_.elevation
              << ", sec = " << source_info_.tv_sec
              << ", usec = " << source_info_.tv_usec;
  }

  {
    kaldi::int32 feat_size, mask_size;
    kaldi::Matrix<kaldi::BaseFloat> feats, masks;
    
    while(socket_.available() < sizeof(feat_size)){
#ifdef _MSC_VER
      ::Sleep((DWORD)1);
#else
      usleep(1000);
#endif
    }
    socket_.receiveBytes(&feat_size, sizeof(feat_size));
    feat_dim_ = feat_size / sizeof(float);
    KALDI_LOG << "dim = " << feat_dim_;
    // Terminated data received. Table.6.89
    if(feat_dim_ == 0){
      KALDI_WARN << "The first frame received, but it has requested terminate.";
      input_finished_ = true;
      return;
    }

    feats.Resize(50, feat_dim_); // The reserve area will be need about 50 frames for as is short utterance.
    masks.Resize(1, feat_dim_); // The reserve area is needless, because the mask data will be disposal finally.
                                // If you implements mask processing, change reserved area to 50.
    frame_size_ = ((sizeof(kaldi::int32) + sizeof(float) * feat_dim_) * 2);

    int remain_size = frame_size_ - sizeof(kaldi::int32); // already read a feat_size
    while(socket_.available() < remain_size){
#ifdef _MSC_VER
      ::Sleep((DWORD)1);
#else
      usleep(1000);
#endif
    }
    socket_.receiveBytes(feats.RowData(0), feat_size);
    socket_.receiveBytes(&mask_size, sizeof(mask_size));
    // The dimensions between the feature and the mask are mismatch.
    if(feat_size != mask_size){
      KALDI_WARN << "The dimensions between the feature and the mask are mismatch.";
      input_finished_ = true;
      return;
    }
    socket_.receiveBytes(masks.RowData(0), mask_size);
    features_ = feats;
    num_frames_ = 1;
  }
}

OnlineFeatureMfcnet::~OnlineFeatureMfcnet() {
  socket_.close();
}

kaldi::int32 OnlineFeatureMfcnet::Dim() const
{
  return feat_dim_;
}

kaldi::int32 OnlineFeatureMfcnet::NumFramesReady() const
{
  return num_frames_;
}

bool OnlineFeatureMfcnet::IsLastFrame(kaldi::int32 frame) const
{
  ////if (input_finished_) { KALDI_WARN << "\nframe=" << frame << ", num_frames_=" << num_frames_ << ", features_.NumRow()=" << features_.NumRows(); }
  return (input_finished_ ? ((frame + 1) == num_frames_) : false);
}

kaldi::BaseFloat OnlineFeatureMfcnet::FrameShiftInSeconds() const
{
  return 0.01f;
}

void OnlineFeatureMfcnet::GetFrame(kaldi::int32 frame, kaldi::VectorBase<kaldi::BaseFloat> *feat)
{
  Receive();
  KALDI_ASSERT(frame >= 0 && frame < num_frames_);
  KALDI_ASSERT(feat->Dim() == Dim());
  feat->CopyFromVec(features_.Row(frame));
}

void OnlineFeatureMfcnet::Receive()
{ // feat_dim_ == Dim()
  if(input_finished_){ return; }
  if(socket_.available() == 0){
    if(socket_.poll(Poco::Timespan(), Poco::Net::Socket::SELECT_READ) && socket_.available() == 0){
      KALDI_WARN << "terminated frame was not received, but socket are timeout.";
      input_finished_ = true;
    }
    return;
  }
  int available_read_size = socket_.available();
  if(available_read_size < sizeof(kaldi::int32)){ return; }
  int available_frame_count = (available_read_size + frame_size_ - sizeof(kaldi::int32)) / frame_size_; // always extra frame for terminated frame
  //int available_frame_count = available_read_size / frame_size_;
  //if(available_read_size - available_frame_count * frame_size_ == sizeof(kaldi::int32)){ available_frame_count++; }
  //available_frame_count = 1; // Always one shot.
                               // This line was wrote for the simulate same as speechrecog behavior,
                               // but this is maybe original bug.
                               // Because variable assigned value, but always meaningless by overwritten 1.

  kaldi::int32 correct_feat_size = feat_dim_ * sizeof(float);
  kaldi::int32 feat_size, mask_size;
  kaldi::Matrix<kaldi::BaseFloat> feats, masks;
  feats.Resize(available_frame_count, feat_dim_);
  masks.Resize(available_frame_count, feat_dim_);
  bool empty = false;

//#if (KALDI_DOUBLEPRECISION != 0) is not supported. if you use double, received socket data and goto copy feat matrix
  int remain_size = frame_size_ - sizeof(kaldi::int32);
  for(int i = 0; i < available_frame_count; i++){
    socket_.receiveBytes(&feat_size, sizeof(feat_size));
    // terminated data received. Table.6.89
    if(feat_size == 0){
      // In the kaldi::Matrix .Resize method,
      // assertion occurred when the matrix size was changed to 0.
      if(i != 0){
        feats.Resize(i, feat_dim_);
        masks.Resize(i, feat_dim_);
      }
      else{
        empty = true;
      }
      available_frame_count = i;
      input_finished_ = true;
      break; // loop end, but use already received frame.
    }
    if(feat_size != correct_feat_size){
      KALDI_WARN << "source_id = " << source_info_.source_id
                 << ", dim = " << feat_size / sizeof(float) << " (!= " << feat_dim_ << ")";
      if(i != 0){
        feats.Resize(i, feat_dim_);
        masks.Resize(i, feat_dim_);
      }
      else{
        empty = true;
      }
      available_frame_count = i;
      input_finished_ = true;
      break; // loop end, but use already received frame.
    }
    while(socket_.available() < remain_size){ // If feat_size != 0 and the data not received yet, we waiting features data.
#ifdef _MSC_VER
      ::Sleep((DWORD)1);
#else
      usleep(1000);
#endif
    }
    socket_.receiveBytes(feats.RowData(i), correct_feat_size);
    socket_.receiveBytes(&mask_size, sizeof(mask_size));
    if(mask_size != correct_feat_size){
      KALDI_WARN << "source_id = " << source_info_.source_id
                 << ", dim = " << mask_size / sizeof(float) << " (!= " << feat_dim_ << ")";
      if(i != 0){
        feats.Resize(i, feat_dim_);
        masks.Resize(i, feat_dim_);
      }
      else{
        empty = true;
      }
      available_frame_count = i;
      input_finished_ = true;
      break; // loop end, but use already received frame.
    }
    socket_.receiveBytes(masks.RowData(i), correct_feat_size);
  }
  KALDI_VLOG(3) << "source_id = " << source_info_.source_id
                << ", received frames = " << available_frame_count;

  if(!empty){
    kaldi::int32 new_num_frames = num_frames_ + feats.NumRows();
    const float increase_ratio = 1.5f;
    if(new_num_frames > features_.NumRows()){
      kaldi::int32 new_num_rows = std::max<kaldi::int32>(new_num_frames, features_.NumRows() * increase_ratio);
      features_.Resize(new_num_rows, feat_dim_, kCopyData);
    }
    features_.Range(num_frames_, feats.NumRows(), 0, feat_dim_).CopyFromMat(feats);
    num_frames_ = new_num_frames;
  }
}

} // namespace kaldi
